#include "fdsolver.h"


void fdsolver(double salt, int * __restrict__ idpos, double * __restrict__ dbcorrection, 
	double * __restrict__ even_divisor, double * __restrict__ odd_divisor, int * __restrict__ charge_index, 
	double * __restrict__ qval, double * __restrict__ even_phi, double * __restrict__ odd_phi,
	int number_dielectrics1, int number_dielectrics2, int number_charges1, int number_charges2, 
	double omega, int nx, int ny, int nz, double max_error, int max_iterations, string calculation_mode)
{

	// ITERATION VARIABLES
    double loop_error;
    const int iter_convergence_check = 10;

    //  VARIABLES FOR THE PROFILING
    struct timespec time0, time1;

    double timeInit = 0.0;
    double timeLaplace = 0.0;
    double timeCharges = 0.0;
    double timeBoundaries = 0.0;
    double timeDielectricBoundaries = 0.0;
    double timeOthers = 0.0;
    double timeTotal = 0.0;
    double timeIteration = 0.0;
    double timeMpi = 0.0;
    int totalIterations = 1;

    initProfiler(time0);

    bool usingDivisor;
    if (salt==0.0)
    	usingDivisor = false;
    else
    	usingDivisor = true;

    const double c1 = omega/6.0;
    const double c2 = 1.0 - omega;

    // Number of even and odd points
    int  ntp,nep,nop;
    ntp = nx*ny*nz;

    if (nz%2==0)
    {
	    nep = ntp/2;
	    nop = nep;
    }
    else
    {
	    nop = (ntp-1)/2;
	    nep = nop+1;    	
    }

	CUDA_profile cudaProfile = {0.0};

	double *even_dielectric_boundary_correction = dbcorrection;
	double *odd_dielectric_boundary_correction = dbcorrection + number_dielectrics1*6;
	int *even_dielectric_boundaries = idpos;
	int *odd_dielectric_boundaries = idpos + number_dielectrics1;

	int *even_charges_index = charge_index;
	int *odd_charges_index = charge_index + number_charges1;
	
	double  *even_bb = qval;
	double  *odd_bb = qval + number_charges1;

	const int number_even_charges = number_charges1;
	const int number_odd_charges = number_charges2 - number_charges1;
	const int number_even_dielectric_boundaries = number_dielectrics1;
	const int number_odd_dielectric_boundaries = number_dielectrics2 - number_dielectrics1;

    if (calculation_mode.compare("gpu") == 0)
    {

    	cu_getDeviceInfo();

		cu_init(0);

		// Init memory
		cu_poisson_initPhi(nx, ny, nz, cudaProfile);

		// Copy initial data
		cu_poisson_copyPhiToGPU(nx, ny, nz, even_phi, odd_phi, cudaProfile);

		// Copy charges
		cu_poisson_initCopyCharges(nx, ny, nz, 1, number_even_charges, even_charges_index, even_bb,
			number_odd_charges, odd_charges_index, odd_bb, cudaProfile);

		// Copy dielectric boundaries
		cu_poisson_initCopyDielectricBoundaries(nx, ny, nz,
			number_even_dielectric_boundaries, even_dielectric_boundaries, even_dielectric_boundary_correction,
			number_odd_dielectric_boundaries, odd_dielectric_boundaries, odd_dielectric_boundary_correction,
			1, cudaProfile);

		// Copy divisor
    	if (usingDivisor)
    		cu_poisson_initCopyDivisor(nep, nop, even_divisor, odd_divisor, cudaProfile);

		cu_poisson_initComparePhi(ntp,ntp-1,cudaProfile);

    }

    

	profiler(time0,time1,timeInit);


	// **************************************
    // HERE BEGINS THE MAIN ITERATION
    // **************************************

    printf("Starting to iterate ********** \n");

	int nn;

    for (nn=1; nn<max_iterations; nn++)
    {

		// if "abort.now" exists, the iteratios is aborted
		ifstream my_file("abort.now");
		if (my_file.good())
		{
			printf("Found abort.now file. Aborting\n");
			my_file.close();
			break;
		}

		initProfiler(time0);

		/** check if we have achieved convergence on the charged grid points */
		if (nn%iter_convergence_check == 0)
		{			
			    // Save boundaries
			    cu_poisson_runSaveLinearConvergenceCheck(nx, ny, nz, number_even_charges, 0, cudaProfile);

			    cu_poisson_runSaveLinearConvergenceCheck(nx, ny, nz, number_odd_charges, 1, cudaProfile);				
		}
		
		profiler(time0,time1,timeOthers);


	    /** iterate over odd grid points to solve laplace only with cuda */

	    if (usingDivisor)
	    {
		    cu_poisson_runBoltzmann(nx, ny, nz, 1, 1, c2, cudaProfile);
	    }
	    else
	    {
		    cu_poisson_runLaplace(nx, ny, nz, 1, c1, c2, cudaProfile);
	    }

		profiler(time0,time1,timeLaplace);


		/** MAKE CORRECTION ON ODD DIELECTRIC BOUNDARY POINTS */
	    cu_poisson_runDielectricBoundaries( nx, ny, nz, number_odd_dielectric_boundaries, 1, cudaProfile);


		profiler(time0,time1,timeDielectricBoundaries);

		/** MAKE CORRECTION ON ODD CHARGED POINTS */
		cu_poisson_runCharges(number_odd_charges, 1, cudaProfile);

		profiler(time0,time1,timeCharges);


		/** iterate over EVEN grid points to solve laplace only */

		if (usingDivisor)
		{
			cu_poisson_runBoltzmann(nx, ny, nz, 1, 0, c2, cudaProfile);
		}
		else
		{
			cu_poisson_runLaplace(nx, ny, nz, 0, c1, c2, cudaProfile);
		}

		profiler(time0,time1,timeLaplace);

		/** make corrections on even boundary grid points */
	    cu_poisson_runDielectricBoundaries( nx, ny, nz,	number_even_dielectric_boundaries, 0, cudaProfile);

		profiler(time0,time1,timeDielectricBoundaries);

		/** correct charged even grid points */
	    cu_poisson_runCharges(number_even_charges, 0, cudaProfile);

		profiler(time0,time1,timeCharges);


		/** check if convergence has been achieved */

		if (nn%iter_convergence_check == 0) /* test convergence */
		{
		    // Get max error
			double maxErrorEven = cu_poisson_runLinearCheckConvergence( nx, ny, nz, number_even_charges, 0, cudaProfile);
			double maxErrorOdd = cu_poisson_runLinearCheckConvergence( nx, ny, nz, number_odd_charges, 1, cudaProfile);
		
			loop_error = max(maxErrorEven, maxErrorOdd);

		    if (loop_error < max_error)
		    {
				printf("Convergence obtained after %d iterations\n",nn);
				break;
		    }
		    else
		    {
				printf("Iteration %d error %f\n", nn, loop_error);
		    }
		}
		

		profiler(time0,time1,timeOthers);

    }
    /** main loop ended */


    /** get back the results from the gpu */
	cu_poisson_copyPhiFromGPU(nx, ny, nz, even_phi, odd_phi, cudaProfile);
 
    
    if (nn == max_iterations)
    {
		printf("  convergence was never obtained\n");
    }

    /** print profiling */
    totalIterations = nn;
    timeTotal =
	    timeInit + timeLaplace +
	    timeCharges + timeBoundaries +
	    timeDielectricBoundaries + timeOthers + timeMpi;
    timeIteration =
	    (timeLaplace + timeCharges +
		timeBoundaries + timeDielectricBoundaries +
		timeOthers + timeMpi) / totalIterations;

	printf("  Total execution time  %f9\n",timeTotal);
    printf("  Specific execution time (seconds):\n");
    printf("    Init                %f9\n",timeInit);
    if (usingDivisor)
    {
    	printf("    Boltzmann             %f9\n",timeLaplace);
    }
    else
    {
    	printf("    Laplace             %f9\n",timeLaplace);
    }
    printf("    Adjust charges      %f9\n",timeCharges);
    printf("    Sum boundaries      %f9\n",timeBoundaries);
    printf("    Adjust diel. bound. %f9\n",timeDielectricBoundaries);
    printf("    Others              %f9\n",timeOthers);
    printf("    Total               %f9\n",timeTotal);
    printf("    Per iteration       %f9\n",timeIteration);
    printf("    On MPI              %f9\n",timeMpi);		


	cu_poisson_destroyMemory(usingDivisor, 1, cudaProfile);

	timeTotal =
		cudaProfile.memoryAllocation +
		cudaProfile.memoryRelease +
		cudaProfile.textureBind +
		cudaProfile.textureUnbind +
		cudaProfile.host2device +
		cudaProfile.device2host +
		cudaProfile.device2device +
		cudaProfile.prepareData +
		cudaProfile.kernelLaplace +
		cudaProfile.kernelCharges +
		cudaProfile.kernelBoundaries +
		cudaProfile.kernelDielectricBoundaries +
		cudaProfile.kernelSaveDielectricBoundaries +
		cudaProfile.kernelCompareDielectricBoundaries +
		cudaProfile.others;

	printf("  CUDA time (seconds):\n");
	printf("    Memory Allocation   %f9\n",cudaProfile.memoryAllocation);
	printf("    Memory Release      %f9\n",cudaProfile.memoryRelease);
	printf("    Texture bind        %f9\n",cudaProfile.textureBind);
	printf("    Texture unbind      %f9\n",cudaProfile.textureUnbind);
	printf("    Host to Device      %f9\n",cudaProfile.host2device);
	printf("    Device to Host      %f9\n",cudaProfile.device2host);
	printf("    Device to Device    %f9\n",cudaProfile.device2device);
	printf("    Prepare Data        %f9\n",cudaProfile.prepareData);
	printf("    Kernel Laplace      %f9\n",cudaProfile.kernelLaplace);
	printf("    Kernel Charges      %f9\n",cudaProfile.kernelCharges);
	printf("    Kernel Boundaries   %f9\n",cudaProfile.kernelBoundaries);
	printf("    Kernel Diel. Bound. %f9\n",cudaProfile.kernelDielectricBoundaries);
	printf("    Kernel Save Bound.  %f9\n",cudaProfile.kernelSaveDielectricBoundaries);
	printf("    Kernel Comp. Bound. %f9\n",cudaProfile.kernelCompareDielectricBoundaries);
	printf("    Others              %f9\n",cudaProfile.others);
	printf("    Total               %f9\n",timeTotal);
    

	cu_finish();
   
    return;
}

